
from sklearn.preprocessing import MinMaxScaler
import pandas as pd

def dummify(data : pd.DataFrame, columns= None, prefix_separator = "-"):
    '''
    dummifing function
    data : original data with categorical and continuous variables
    prefix_sep  mostly uses '_' as prefix separator. Can be modified
    columns : list of columns that are categorical
    '''
    excluded_variables = [continuous for continuous in data.columns if continuous not in columns]
    data1 = data[excluded_variables].copy() # we will concat this data with the dummified data
    
    list_of_series = [data[cols] for cols in columns]
    dummified = list(map(lambda x, y : pd.get_dummies(data=y, prefix=x, prefix_sep=prefix_separator), columns, list_of_series))
    
    dummified = pd.concat([data1, *dummified], axis=1)
    
    return dummified


def undummify(data :pd.DataFrame, columns :list, prefix_separator="-"):
    '''
    This function changes tabular data's categorical variable's one-hot dummy encoding to the original form
    data : original tabular data
    columns : a list of names of categorical values
    prefix_sep : prefix separator. Mostly is "-"
    '''
    # cols2collapse = {
    #     item.split(prefix_separator)[0]: (prefix_separator in item) for item in data.columns
    # }
    
    continuous_list = [cont for cont in data.columns if cont.split(prefix_separator)[0] not in columns]
    data1 = data[continuous_list].copy() # data to concat later

    undummified_list  = []
    for col in columns:
        undummified = data.filter(like=col).idxmax(axis=1).apply(lambda x : x.split(prefix_separator, maxsplit=1)[1]).rename(col)
        undummified_list.append(undummified) 
    
    recovered_data = pd.concat([data1, *undummified_list], axis=1)
    return recovered_data



def do_scaling(data : pd.DataFrame, columns : list):
    '''
    scaling function for continous variable. 
    This function returns the scaled data and the scaler itself for future use.
    data : the data you want to change
    columns : continuous columns
    '''
    scaler = MinMaxScaler(feature_range=(-1,1))
    not_scaling_columns = [cols for cols in data.columns if cols not in columns ]
    data1 = data[not_scaling_columns].copy()
    
    scaler.fit(data[columns].copy())
    scaled = scaler.transform(data[columns].copy())
    scaled= pd.DataFrame(scaled, columns=columns)
    
    scaled_data = pd.concat([scaled,data1], axis=1)
    return scaled_data, scaler


def reverse_scaling(data : pd.DataFrame, columns : list, scaler : MinMaxScaler):
    '''
    reverse the scaled data to original form
    data : scaled data
    columns : continuous columns - should be the same as the original scaling process
    scaler : Minmax scaler from do_scaling function
    '''
    excluded_cols = [col for col in data.columns if col not in columns]
    
    data1 = data[excluded_cols].copy()
    inversed = scaler.inverse_transform(data[columns])
    inversed = pd.DataFrame(inversed, columns=columns)
    
    reversed_data = pd.concat([data1, inversed], axis=1)
    return reversed_data




__version__ = '0.4.0'

from .bounded_laplace_mechanism import *
from .functions import *
from .laplace_mechanism import *
from .original_timedp import *
from .preprocessing import *
from .timedp_algorithm import *


import numpy as np

def laplace_mechanism(value : float, sensitivity: float, epsilon:float, seed=0):
    if epsilon == 0 :
        return value
    np.random.seed(seed)
    scale = sensitivity/epsilon
    noise = np.random.laplace(0, size=1, scale=scale)
    return value + noise.item()




import numpy as np
from scipy import interpolate

def make_extended_vector(vector : np.array):
    '''
    This function takes a given function and outputs a extended version of the vector
    The extended values are the linear interpolated values
    '''
    length_of_timestep = len(vector)
    xs = [2*x+1 for x in range(0,length_of_timestep)]
    new_xs = [x for x in range(0,2*length_of_timestep)]
    
    f = interpolate.interp1d(xs, vector, fill_value='extrapolate')
    
    return f(new_xs)


def make_coordinate_list(x_vector : list, y_vector : list, step: int):
    coordinate_list = [(x,y) for x, y in zip(x_vector, y_vector)][::step]
    return coordinate_list


def calculate_gradient(two_points:tuple, dx=1):
    '''
    This function calculates the gradient of a two point.
    The precondition is that the dx of a given two points is 1. But this can be changed.
    two_points : (x_0, x_1) 
    '''
    if dx <= 0 : 
        raise ValueError('dx should be larger than 0')
    
    x0, x1 = two_points
    grad = (x1 - x0)/dx
    return grad


def give_calculated_gradient_list(extended_vector:list, step=2):
    '''
    This function takes the extended vector as its input and calculates the gradient.
    It outputs the gradient for every point.
    step : step size to calculate the gradient. default : 2
    '''
    number_of_chunks = int(len(extended_vector) / 2)
    chunks = [(extended_vector[num*2],extended_vector[num*2+1]) for num in range(0,number_of_chunks)]
    return list(map(calculate_gradient,chunks))


def calculate_linear_function_form(coordinate: tuple, gradient: float):
    '''
    This function calculates the linear function parameters.
    coordinate : (x,y)
    y = ax + b
    '''
    x, y = coordinate
    b = y - (x*gradient)
    return gradient, b



#!/bin/python
import numpy as np

def cal_C(value, D, b):
    l, u = D
    if b < 0.001 :
        b = 0.001 # set lower bound of b
    return 1 - (1/2)*(np.exp((l-value)/b) + np.exp((value-u)/b))


def boundedlaplacemechanism(value, D: tuple, b, epsilon, delta, seed=0): 
    '''
    density function of Wq(bounded laplace mechanism)
    D : the support for x(value)
    b : variance -> requires checkup. 
    sensitivity : the minimal difference 
    '''
    # if the epsilon is 0, return the original value
    if epsilon == 0 :
        return value
    
    if type(D) == None :
        return 0
    # assign boundaries
    l, u  = D
    
    # Theorem 4.4(Fixed Point)
    dQ = abs(u - l)
    
    #Definition3.4
    # calculate Cl, Cl_dQ 
    Cl = cal_C(l, D, b)
    Cl_dQ = cal_C(l + dQ, D, b)
    
    dC = Cl_dQ / Cl

    if b < dQ/(epsilon - np.log(dC) - np.log(1 - delta)) :
        b = dQ/(epsilon - np.log(dC) - np.log(1 - delta))

    # Here we do not calculate Cq 
    # Instead we calculate the function and then sum the output calculations
    theRange = np.linspace(l, u, num = 10**4)
    e_term = np.exp(-(abs(theRange - value)/b)) 
    output_calculation  = (1/(2*b))*(e_term)
    
    # sum the output calculation and normalize
    theSum = output_calculation.sum()
    density = output_calculation/theSum 
    
    # the function will output a randomized output!
    np.random.seed(seed=seed)
    randomized_output = np.random.choice(theRange, 1, p=density)

    return randomized_output.item()



import pandas as pd
import numpy as np
import pickle
from sklearn.preprocessing import MinMaxScaler
from copy import deepcopy
from multiprocessing import Pool, freeze_support
import os,sys


def get_differential_privacy_value(value, epsilon, seed=0):
    
    def pdf(x):
        b = 2 / (epsilon)
        c = 1 - 0.5 * (np.exp(-(value+1)/b) + np.exp(-(1 - value)/b))
        return (1 / (b * c * 2)) * np.exp(-np.absolute(x - value)/b)

    elements = np.linspace(-1, 1, 10**4)
    probabilities = pdf(elements)
    probabilities /= np.sum(probabilities)
    return np.random.choice(elements, size=1, p=probabilities.reshape(-1)).item()


def timeseries_dp(timeseries, epsilon):
    scaler = MinMaxScaler(feature_range=(-1,1))
    
    S = []
    
    for val_idx, value in enumerate(timeseries):
        if val_idx == 0 :
            ranges = np.array([value,timeseries[val_idx+1]]).reshape(-1,1)
            scaler.fit(ranges)
            scaled_value = scaler.transform(ranges)[0]
            dp_value = get_differential_privacy_value(scaled_value, epsilon) # 여기를 나중에 바꿔줄 것임
            # synthesized value : v'
            dp_value = np.array(dp_value).reshape(-1,1)
            print(dp_value)
            syn_value = scaler.inverse_transform(dp_value)
            S.append(syn_value.item())
    
    ## 2. get all values after index 0
        elif val_idx == len(timeseries)-1 :
            ranges = np.array([value,timeseries[val_idx-1]]).reshape(-1,1)
            scaler.fit(ranges)
            scaled_value = scaler.transform(ranges)[0]
            dp_value = get_differential_privacy_value(scaled_value, epsilon)
            dp_value = np.array(dp_value).reshape(-1,1)
            syn_value = scaler.inverse_transform(dp_value)
            S.append(syn_value.item())
            
        else :
            v_1, v_3 = timeseries[val_idx-1], timeseries[val_idx+1] # value index +1 을 v_3로 표현
            if (v_1 < value < v_3) | (v_1 > value > v_3) :
                ranges = np.array([v_1, v_3]).reshape(-1,1)
                scaler.fit(ranges)
                scaled_value = scaler.transform(np.array(value).reshape(-1,1))
                dp_value = get_differential_privacy_value(scaled_value, epsilon)
                dp_value = np.array(dp_value).reshape(-1,1)
                syn_value = scaler.inverse_transform(dp_value)
                S.append(syn_value.item())
                
            
            elif (v_1 < value) & (v_3 < value):
                _, second, first = sorted([v_1, value, v_3])
                ranges = np.array([first, second]).reshape(-1,1)
                scaler.fit(ranges)
                scaled_value = scaler.transform(np.array(value).reshape(-1,1))
                dp_value = get_differential_privacy_value(scaled_value, epsilon)
                dp_value = np.array(dp_value).reshape(-1,1)
                syn_value = scaler.inverse_transform(dp_value)
                S.append(syn_value.item())
                
            elif (v_1 > value) & (v_3 > value) :
                third, second, _ = sorted([v_1, value, v_3])
                ranges = np.array([third, second]).reshape(-1,1)
                scaler.fit(ranges)
                scaled_value = scaler.transform(np.array(value).reshape(-1,1))
                dp_value = get_differential_privacy_value(scaled_value, epsilon)
                dp_value = np.array(dp_value).reshape(-1,1)
                syn_value = scaler.inverse_transform(dp_value)
                S.append(syn_value.item())
                
            else :
                if (value == v_1 == v_3) :
                    # syn_value = get_differential_privacy_value(value, epsilon)
                    syn_value = 0
                    S.append(syn_value)
                else : 
                    val1, val2 = list(set([value, v_1, v_3]))
                    ranges = np.array([val1, val2]).reshape(-1,1)
                    scaler.fit(ranges)
                    scaled_value = scaler.transform(np.array(value).reshape(-1,1))
                    dp_value = get_differential_privacy_value(scaled_value, epsilon)
                    dp_value = np.array(dp_value).reshape(-1,1)
                    syn_value = scaler.inverse_transform(dp_value)
                    S.append(syn_value.item())
    return S





from syndp.functions import *
from syndp.bounded_laplace_mechanism import *
from syndp.laplace_mechanism import *

class TimeDP:
    
    def __init__(self, epsilon, delta, mechanism_type : str, sensitivity=0.1, seed=0):
        '''
        This class is a noise giving class. It takes the gradient and calculates new synthezied series data
        you can choose two types of mechanism. Original laplace mechanism or Bounded Laplace Mechanism
        mechanism_type : 1) laplace, 2) bouned_laplace
        '''
        self.epsilon = epsilon
        self.delta = delta 
        self.mechanism_type = mechanism_type
        self.sensitivity = sensitivity
        self.mechanism = self._dp_mechanism()
        self.seed = seed
    
    def _dp_mechanism(self):
        if self.mechanism_type == 'laplace':
            return laplace_mechanism
        else :
            return boundedlaplacemechanism
    
    def calculate_dp_value(self, val, D = None, seed=0):
        '''
        requires value(val) and sensitivity(sens)
        '''
        if self.mechanism_type == 'laplace':
            return self.mechanism(value=val, sensitivity=self.sensitivity, epsilon=self.epsilon, seed=seed)
        else :
            self.sensitivity = 9999 # change to 9999
            return self.mechanism(value=val, D = D, b=self.sensitivity, epsilon=self.epsilon, delta = self.delta, seed=self.seed)


class Vector_creator:
    
    def __init__(self, vector: np.array, timedp: object):
        self.vector = vector
        self.vector_length = len(vector)
        self.timedp = timedp
        self.mechanism_type =timedp.mechanism_type
        self.extended_vector = self.make_extended_vector()
        self.coordinates = self.make_coordinates()
        self.gradient_list = self.prepare_for_gradient()
        self.new_gradients = self.make_new_gradient()
        self.new_function_forms = self.calculate_function_form()
        self.new_vector = self.make_new_vector()
        
    def make_extended_vector(self):
        '''
        makes extended vector
        '''
        return make_extended_vector(vector=self.vector)
    
    def make_coordinates(self):
        '''
        makes x and y coordinate list 
        '''
        extend_vector_length = self.vector_length * 2
        xs = [x for x in range(0, extend_vector_length)]
        return make_coordinate_list(xs, self.extended_vector, 2)
    
    
    def prepare_for_gradient(self):
        return give_calculated_gradient_list(self.extended_vector, step=2)
    
    def create_boundary(self, gradient):
        if gradient < 0 :
            return -9999, 0
        elif gradient > 0 :
            return 0, 9999
        else :
            return -1, 1
            
    def create_boundary_list(self):
        
        # print('creating boundary list')
        
        return list(map(self.create_boundary, self.gradient_list))
    
    def make_new_gradient(self):
        
        # print('created boundary list and making new gradients..')
        
        if self.mechanism_type == 'laplace':
            
            seeds = [i+self.timedp.seed for i in range(0, len(self.gradient_list))]
            return list(map(lambda x, y: self.timedp.calculate_dp_value(val = x, seed=y), self.gradient_list, seeds))
        else :
            boundary_list = self.create_boundary_list()
            return list(map(lambda x, y : self.timedp.calculate_dp_value(val=x, D=y, seed=self.timedp.seed), self.gradient_list, boundary_list))
    
    def calculate_function_form(self):
        '''
        list of (gradient and b)
        '''
        return list(map(lambda x, y : calculate_linear_function_form(x,y), self.coordinates, self.new_gradients))
    
    def make_new_value_based_on_new_gradient(self, gradient, b, x):
        y = x*gradient + b
        return y
    
    def make_new_vector(self):
        extend_vector_length = self.vector_length * 2
        xs = [x for x in range(0, extend_vector_length)][1::2]
        return [self.make_new_value_based_on_new_gradient(g,b,x) for (g, b), x, in zip(self.new_function_forms, xs)]
        
        
        



import setuptools



setuptools.setup(
        name = "syndp",
        version = "0.4.0",
        author = "Won Seok Jang",
        author_email = "memy85@naver.com",
        description = "Syndp is data synthesis package that leverages local differential privacy and various synthesis methods",
        long_description_content_type = "text/markdown",
        url = "https://github.com/memy85/Syndp",
        install_requires = ['pandas','numpy','matplotlib','sklearn'],
        classifiers = [
            "Programming Language :: Python :: 3.8",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
            ],
        package_dir={"":"src"},
        packages=setuptools.find_packages(where="src"),
        python_requires=">=3.7"
        ) 
